@@ -24,75 +24,121 @@
  * THE SOFTWARE.
  */
 
+#include "py/builtin.h"
+
+#if MICROPY_PY_BUILTINS_FLOAT && MICROPY_PY_MATH
+
 #include <math.h>
 
-#include "misc.h"
-#include "mpconfig.h"
-#include "qstr.h"
-#include "obj.h"
-#include "builtin.h"
-
-#if MICROPY_ENABLE_FLOAT && MICROPY_ENABLE_MOD_MATH
+/// \module math - mathematical functions
+///
+/// The `math` module provides some basic mathematical funtions for
+/// working with floating-point numbers.
 
 //TODO: Change macros to check for overflow and raise OverflowError or RangeError
 #define MATH_FUN_1(py_name, c_name) \
-    mp_obj_t mp_math_ ## py_name(mp_obj_t x_obj) { return mp_obj_new_float(MICROPY_FLOAT_C_FUN(c_name)(mp_obj_get_float(x_obj))); } \
+    STATIC mp_obj_t mp_math_ ## py_name(mp_obj_t x_obj) { return mp_obj_new_float(MICROPY_FLOAT_C_FUN(c_name)(mp_obj_get_float(x_obj))); } \
     STATIC MP_DEFINE_CONST_FUN_OBJ_1(mp_math_## py_name ## _obj, mp_math_ ## py_name);
 
 #define MATH_FUN_2(py_name, c_name) \
-    mp_obj_t mp_math_ ## py_name(mp_obj_t x_obj, mp_obj_t y_obj) { return mp_obj_new_float(MICROPY_FLOAT_C_FUN(c_name)(mp_obj_get_float(x_obj), mp_obj_get_float(y_obj))); } \
+    STATIC mp_obj_t mp_math_ ## py_name(mp_obj_t x_obj, mp_obj_t y_obj) { return mp_obj_new_float(MICROPY_FLOAT_C_FUN(c_name)(mp_obj_get_float(x_obj), mp_obj_get_float(y_obj))); } \
     STATIC MP_DEFINE_CONST_FUN_OBJ_2(mp_math_## py_name ## _obj, mp_math_ ## py_name);
 
 #define MATH_FUN_1_TO_BOOL(py_name, c_name) \
-    mp_obj_t mp_math_ ## py_name(mp_obj_t x_obj) { return MP_BOOL(c_name(mp_obj_get_float(x_obj))); } \
+    STATIC mp_obj_t mp_math_ ## py_name(mp_obj_t x_obj) { return MP_BOOL(c_name(mp_obj_get_float(x_obj))); } \
     STATIC MP_DEFINE_CONST_FUN_OBJ_1(mp_math_## py_name ## _obj, mp_math_ ## py_name);
 
 #define MATH_FUN_1_TO_INT(py_name, c_name) \
-    mp_obj_t mp_math_ ## py_name(mp_obj_t x_obj) { return mp_obj_new_int((machine_int_t)MICROPY_FLOAT_C_FUN(c_name)(mp_obj_get_float(x_obj))); } \
+    STATIC mp_obj_t mp_math_ ## py_name(mp_obj_t x_obj) { mp_int_t x = MICROPY_FLOAT_C_FUN(c_name)(mp_obj_get_float(x_obj)); return mp_obj_new_int(x); } \
     STATIC MP_DEFINE_CONST_FUN_OBJ_1(mp_math_## py_name ## _obj, mp_math_ ## py_name);
 
 // These are also used by cmath.c
+/// \constant e - base of the natural logarithm
 const mp_obj_float_t mp_math_e_obj = {{&mp_type_float}, M_E};
+/// \constant pi - the ratio of a circle's circumference to its diameter
 const mp_obj_float_t mp_math_pi_obj = {{&mp_type_float}, M_PI};
 
+/// \function sqrt(x)
+/// Returns the square root of `x`.
 MATH_FUN_1(sqrt, sqrt)
+/// \function pow(x, y)
+/// Returns `x` to the power of `y`.
 MATH_FUN_2(pow, pow)
+/// \function exp(x)
 MATH_FUN_1(exp, exp)
+/// \function expm1(x)
 MATH_FUN_1(expm1, expm1)
+/// \function log(x)
 MATH_FUN_1(log, log)
+/// \function log2(x)
 MATH_FUN_1(log2, log2)
+/// \function log10(x)
 MATH_FUN_1(log10, log10)
+/// \function cosh(x)
 MATH_FUN_1(cosh, cosh)
+/// \function sinh(x)
 MATH_FUN_1(sinh, sinh)
+/// \function tanh(x)
 MATH_FUN_1(tanh, tanh)
+/// \function acosh(x)
 MATH_FUN_1(acosh, acosh)
+/// \function asinh(x)
 MATH_FUN_1(asinh, asinh)
+/// \function atanh(x)
 MATH_FUN_1(atanh, atanh)
+/// \function cos(x)
 MATH_FUN_1(cos, cos)
+/// \function sin(x)
 MATH_FUN_1(sin, sin)
+/// \function tan(x)
 MATH_FUN_1(tan, tan)
+/// \function acos(x)
 MATH_FUN_1(acos, acos)
+/// \function asin(x)
 MATH_FUN_1(asin, asin)
+/// \function atan(x)
 MATH_FUN_1(atan, atan)
+/// \function atan2(y, x)
 MATH_FUN_2(atan2, atan2)
+/// \function ceil(x)
 MATH_FUN_1_TO_INT(ceil, ceil)
+/// \function copysign(x, y)
 MATH_FUN_2(copysign, copysign)
+/// \function fabs(x)
 MATH_FUN_1(fabs, fabs)
+/// \function floor(x)
 MATH_FUN_1_TO_INT(floor, floor) //TODO: delegate to x.__floor__() if x is not a float
+/// \function fmod(x, y)
 MATH_FUN_2(fmod, fmod)
+/// \function isfinite(x)
 MATH_FUN_1_TO_BOOL(isfinite, isfinite)
+/// \function isinf(x)
 MATH_FUN_1_TO_BOOL(isinf, isinf)
+/// \function isnan(x)
 MATH_FUN_1_TO_BOOL(isnan, isnan)
+/// \function trunc(x)
 MATH_FUN_1_TO_INT(trunc, trunc)
+/// \function ldexp(x, exp)
 MATH_FUN_2(ldexp, ldexp)
+/// \function erf(x)
+/// Return the error function of `x`.
 MATH_FUN_1(erf, erf)
+/// \function erfc(x)
+/// Return the complementary error function of `x`.
 MATH_FUN_1(erfc, erfc)
+/// \function gamma(x)
+/// Return the gamma function of `x`.
 MATH_FUN_1(gamma, tgamma)
+/// \function lgamma(x)
+/// return the natural logarithm of the gamma function of `x`.
 MATH_FUN_1(lgamma, lgamma)
 //TODO: factorial, fsum
 
 // Functions that return a tuple
-mp_obj_t mp_math_frexp(mp_obj_t x_obj) {
+
+/// \function frexp(x)
+/// Converts a floating-point number to fractional and integral components.
+STATIC mp_obj_t mp_math_frexp(mp_obj_t x_obj) {
     int int_exponent = 0;
     mp_float_t significand = MICROPY_FLOAT_C_FUN(frexp)(mp_obj_get_float(x_obj), &int_exponent);
     mp_obj_t tuple[2];
@@ -102,7 +148,8 @@ mp_obj_t mp_math_frexp(mp_obj_t x_obj) {
 }
 STATIC MP_DEFINE_CONST_FUN_OBJ_1(mp_math_frexp_obj, mp_math_frexp);
 
-mp_obj_t mp_math_modf(mp_obj_t x_obj) {
+/// \function modf(x)
+STATIC mp_obj_t mp_math_modf(mp_obj_t x_obj) {
     mp_float_t int_part = 0.0;
     mp_float_t fractional_part = MICROPY_FLOAT_C_FUN(modf)(mp_obj_get_float(x_obj), &int_part);
     mp_obj_t tuple[2];
@@ -113,12 +160,15 @@ mp_obj_t mp_math_modf(mp_obj_t x_obj) {
 STATIC MP_DEFINE_CONST_FUN_OBJ_1(mp_math_modf_obj, mp_math_modf);
 
 // Angular conversions
-mp_obj_t mp_math_radians(mp_obj_t x_obj) {
+
+/// \function radians(x)
+STATIC mp_obj_t mp_math_radians(mp_obj_t x_obj) {
     return mp_obj_new_float(mp_obj_get_float(x_obj) * M_PI / 180.0);
 }
 STATIC MP_DEFINE_CONST_FUN_OBJ_1(mp_math_radians_obj, mp_math_radians);
 
-mp_obj_t mp_math_degrees(mp_obj_t x_obj) {
+/// \function degrees(x)
+STATIC mp_obj_t mp_math_degrees(mp_obj_t x_obj) {
     return mp_obj_new_float(mp_obj_get_float(x_obj) * 180.0 / M_PI);
 }
 STATIC MP_DEFINE_CONST_FUN_OBJ_1(mp_math_degrees_obj, mp_math_degrees);
@@ -151,6 +201,7 @@ STATIC const mp_map_elem_t mp_module_math_globals_table[] = {
     { MP_OBJ_NEW_QSTR(MP_QSTR_copysign), (mp_obj_t)&mp_math_copysign_obj },
     { MP_OBJ_NEW_QSTR(MP_QSTR_fabs), (mp_obj_t)&mp_math_fabs_obj },
     { MP_OBJ_NEW_QSTR(MP_QSTR_floor), (mp_obj_t)&mp_math_floor_obj },
+    { MP_OBJ_NEW_QSTR(MP_QSTR_fmod), (mp_obj_t)&mp_math_fmod_obj },
     { MP_OBJ_NEW_QSTR(MP_QSTR_frexp), (mp_obj_t)&mp_math_frexp_obj },
     { MP_OBJ_NEW_QSTR(MP_QSTR_ldexp), (mp_obj_t)&mp_math_ldexp_obj },
     { MP_OBJ_NEW_QSTR(MP_QSTR_modf), (mp_obj_t)&mp_math_modf_obj },
@@ -166,16 +217,7 @@ STATIC const mp_map_elem_t mp_module_math_globals_table[] = {
     { MP_OBJ_NEW_QSTR(MP_QSTR_lgamma), (mp_obj_t)&mp_math_lgamma_obj },
 };
 
-STATIC const mp_obj_dict_t mp_module_math_globals = {
-    .base = {&mp_type_dict},
-    .map = {
-        .all_keys_are_qstrs = 1,
-        .table_is_fixed_array = 1,
-        .used = ARRAY_SIZE(mp_module_math_globals_table),
-        .alloc = ARRAY_SIZE(mp_module_math_globals_table),
-        .table = (mp_map_elem_t*)mp_module_math_globals_table,
-    },
-};
+STATIC MP_DEFINE_CONST_DICT(mp_module_math_globals, mp_module_math_globals_table);
 
 const mp_obj_module_t mp_module_math = {
     .base = { &mp_type_module },
@@ -183,4 +225,4 @@ const mp_obj_module_t mp_module_math = {
     .globals = (mp_obj_dict_t*)&mp_module_math_globals,
 };
 
-#endif // MICROPY_ENABLE_FLOAT && MICROPY_ENABLE_MOD_MATH
+#endif // MICROPY_PY_BUILTINS_FLOAT && MICROPY_PY_MATH
