@@ -23,5 +23,11 @@
  * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  * THE SOFTWARE.
  */
+#ifndef __MICROPY_INCLUDED_PY_PARSENUM_H__
+#define __MICROPY_INCLUDED_PY_PARSENUM_H__
 
-int mp_parse_num_base(const char *str, uint len, int *base);
+#include "py/mpconfig.h"
+
+mp_uint_t mp_parse_num_base(const char *str, mp_uint_t len, mp_uint_t *base);
+
+#endif // __MICROPY_INCLUDED_PY_PARSENUM_H__
