@@ -24,7 +24,13 @@
  * THE SOFTWARE.
  */
 
-extern const mp_obj_module_t os_module;
 extern const mp_obj_module_t pyb_module;
 extern const mp_obj_module_t stm_module;
-extern const mp_obj_module_t time_module;
+extern const mp_obj_module_t mp_module_uos;
+extern const mp_obj_module_t mp_module_utime;
+extern const mp_obj_module_t mp_module_uselect;
+extern const mp_obj_module_t mp_module_usocket;
+
+// additional helper functions exported by the modules
+
+mp_uint_t mod_time_seconds_since_2000(mp_uint_t year, mp_uint_t month, mp_uint_t date, mp_uint_t hour, mp_uint_t minute, mp_uint_t second);
