@@ -24,6 +24,9 @@
  * THE SOFTWARE.
  */
 
+extern struct _pyb_uart_obj_t *pyb_stdio_uart;
+
 void stdout_tx_str(const char *str);
-void stdout_tx_strn(const char *str, uint len);
+void stdout_tx_strn(const char *str, mp_uint_t len);
+void stdout_tx_strn_cooked(const char *str, mp_uint_t len);
 int stdin_rx_chr(void);
