@@ -33,12 +33,13 @@
   */
 
 /* Includes ------------------------------------------------------------------*/
+
 #include <stdbool.h>
+
 #include "stm32f4xx_hal.h"
 #include "usbd_cdc_msc_hid.h"
 #include "usbd_cdc_interface.h"
 #include "pendsv.h"
-#include "usb.h"
 
 // CDC control commands
 #define CDC_SEND_ENCAPSULATED_COMMAND               0x00
@@ -59,7 +60,7 @@
 /* Private macro -------------------------------------------------------------*/
 /* Private variables ---------------------------------------------------------*/
 
-static uint8_t dev_is_connected = 0; // indicates if we are connected
+static __IO uint8_t dev_is_connected = 0; // indicates if we are connected
 
 static uint8_t UserRxBuffer[APP_RX_DATA_SIZE]; // received data from USB OUT endpoint is stored in this buffer
 static uint16_t UserRxBufCur = 0; // points to next available character in UserRxBuffer
@@ -72,7 +73,7 @@ static uint16_t UserTxBufPtrOutShadow = 0; // shadow of above
 static uint8_t UserTxBufPtrWaitCount = 0; // used to implement a timeout waiting for low-level USB driver
 static uint8_t UserTxNeedEmptyPacket = 0; // used to flush the USB IN endpoint if the last packet was exactly the endpoint packet size
 
-static int user_interrupt_char = VCP_CHAR_NONE;
+static int user_interrupt_char = -1;
 static void *user_interrupt_data = NULL;
 
 /* USB handler declaration */
@@ -152,7 +153,7 @@ static int8_t CDC_Itf_Init(void)
      * may be called before this init function to set these values.
      * This can happen if the USB enumeration occurs after the call to
      * USBD_CDC_SetInterrupt.
-    user_interrupt_char = VCP_CHAR_NONE;
+    user_interrupt_char = -1;
     user_interrupt_data = NULL;
     */
 
@@ -344,7 +345,7 @@ static int8_t CDC_Itf_Receive(uint8_t* Buf, uint32_t *Len) {
 
     uint32_t delta_len;
 
-    if (user_interrupt_char == VCP_CHAR_NONE) {
+    if (user_interrupt_char == -1) {
         // no special interrupt character
         delta_len = *Len;
 
@@ -357,6 +358,8 @@ static int8_t CDC_Itf_Receive(uint8_t* Buf, uint32_t *Len) {
         for (; src < buf_top; src++) {
             if (*src == user_interrupt_char) {
                 char_found = true;
+                // raise exception when interrupts are finished
+                pendsv_nlr_jump(user_interrupt_data);
             } else {
                 if (char_found) {
                     *dest = *src;
@@ -365,12 +368,6 @@ static int8_t CDC_Itf_Receive(uint8_t* Buf, uint32_t *Len) {
             }
         }
 
-        if (char_found) {
-            // raise exception when interrupts are finished
-            user_interrupt_char = VCP_CHAR_NONE;
-            pendsv_nlr_jump(user_interrupt_data);
-        }
-
         // length of remaining characters
         delta_len = dest - Buf;
     }
@@ -399,7 +396,42 @@ void USBD_CDC_SetInterrupt(int chr, void *data) {
     user_interrupt_data = data;
 }
 
-void USBD_CDC_Tx(const char *str, uint32_t len) {
+int USBD_CDC_TxHalfEmpty(void) {
+    int32_t tx_waiting = (int32_t)UserTxBufPtrIn - (int32_t)UserTxBufPtrOut;
+    if (tx_waiting < 0) {
+        tx_waiting += APP_TX_DATA_SIZE;
+    }
+    return tx_waiting <= APP_TX_DATA_SIZE / 2;
+}
+
+// timout in milliseconds.
+// Returns number of bytes written to the device.
+int USBD_CDC_Tx(const uint8_t *buf, uint32_t len, uint32_t timeout) {
+    for (uint32_t i = 0; i < len; i++) {
+        // Wait until the device is connected and the buffer has space, with a given timeout
+        uint32_t start = HAL_GetTick();
+        while (!dev_is_connected || ((UserTxBufPtrIn + 1) & (APP_TX_DATA_SIZE - 1)) == UserTxBufPtrOut) {
+            // Wraparound of tick is taken care of by 2's complement arithmetic.
+            if (HAL_GetTick() - start >= timeout) {
+                // timeout
+                return i;
+            }
+            __WFI(); // enter sleep mode, waiting for interrupt
+        }
+
+        // Write data to device buffer
+        UserTxBuffer[UserTxBufPtrIn] = buf[i];
+        UserTxBufPtrIn = (UserTxBufPtrIn + 1) & (APP_TX_DATA_SIZE - 1);
+    }
+
+    // Success, return number of bytes read
+    return len;
+}
+
+// Always write all of the data to the device tx buffer, even if the
+// device is not connected, or if the buffer is full.  Has a small timeout
+// to wait for the buffer to be drained, in the case the device is connected.
+void USBD_CDC_TxAlways(const uint8_t *buf, uint32_t len) {
     for (int i = 0; i < len; i++) {
         // If the CDC device is not connected to the host then we don't have anyone to receive our data.
         // The device may become connected in the future, so we should at least try to fill the buffer
@@ -433,23 +465,36 @@ void USBD_CDC_Tx(const char *str, uint32_t len) {
             */
         }
 
-        UserTxBuffer[UserTxBufPtrIn] = str[i];
+        UserTxBuffer[UserTxBufPtrIn] = buf[i];
         UserTxBufPtrIn = (UserTxBufPtrIn + 1) & (APP_TX_DATA_SIZE - 1);
     }
 }
 
+// Returns number of bytes in the rx buffer.
 int USBD_CDC_RxNum(void) {
     return UserRxBufLen - UserRxBufCur;
 }
 
-int USBD_CDC_RxGet(void) {
-    // wait for buffer to have at least 1 character in it
-    while (USBD_CDC_RxNum() == 0) {
-        __WFI();
+// timout in milliseconds.
+// Returns number of bytes read from the device.
+int USBD_CDC_Rx(uint8_t *buf, uint32_t len, uint32_t timeout) {
+    // loop to read bytes
+    for (uint32_t i = 0; i < len; i++) {
+        // Wait until we have at least 1 byte to read
+        uint32_t start = HAL_GetTick();
+        while (UserRxBufLen == UserRxBufCur) {
+            // Wraparound of tick is taken care of by 2's complement arithmetic.
+            if (HAL_GetTick() - start >= timeout) {
+                // timeout
+                return i;
+            }
+            __WFI(); // enter sleep mode, waiting for interrupt
+        }
+
+        // Copy byte from device to user buffer
+        buf[i] = UserRxBuffer[UserRxBufCur++];
     }
 
-    // get next character
-    int c = UserRxBuffer[UserRxBufCur++];
-
-    return c;
+    // Success, return number of bytes read
+    return len;
 }
