@@ -33,7 +33,13 @@ print(bytes(3))
 print(bytes([3, 2, 1]))
 print(bytes(range(5)))
 
-def gen():
-    for i in range(4):
-        yield i
-print(bytes(gen()))
+# Make sure bytes are not mistreated as unicode
+x = b"\xff\x8e\xfe}\xfd\x7f"
+print(len(x))
+print(x[0], x[1], x[2], x[3])
+
+# Make sure init values are not mistreated as unicode chars
+# For sequence of known len
+print(bytes([128, 255]))
+# For sequence of unknown len
+print(bytes(iter([128, 255])))
