@@ -1,3 +1,6 @@
+print(int(False))
+print(int(True))
+
 print(int(0))
 print(int(1))
 print(int(+1))
@@ -47,6 +50,7 @@ print(int('0100', 2))
 print(int(' \t 0o12', 8))
 print(int('0o12  \t  ', 8))
 print(int(b"12", 10))
+print(int(b"12"))
 
 
 def test(value, base):
@@ -74,3 +78,6 @@ test('0b2', 2)
 test('0o8', 8)
 test('0xg', 16)
 test('1 1', 16)
+
+# check that we don't parse this as a floating point number
+print(0x1e+1)
