@@ -3,22 +3,22 @@ try:
 except ImportError:
     from _collections import namedtuple
 
-T = namedtuple("Tup", "foo bar")
+T = namedtuple("Tup", ["foo", "bar"])
 # CPython prints fully qualified name, what we don't bother to do so far
 #print(T)
-t = T(1, 2)
-print(t)
-print(t[0], t[1])
-print(t.foo, t.bar)
+for t in T(1, 2), T(bar=1, foo=2):
+    print(t)
+    print(t[0], t[1])
+    print(t.foo, t.bar)
 
-print(len(t))
-print(bool(t))
-print(t + t)
-print(t * 3)
+    print(len(t))
+    print(bool(t))
+    print(t + t)
+    print(t * 3)
 
-print([f for f in t])
+    print([f for f in t])
 
-print(isinstance(t, tuple))
+    print(isinstance(t, tuple))
 
 try:
     t[0] = 200
@@ -27,7 +27,7 @@ except TypeError:
 try:
     t.bar = 200
 except AttributeError:
-    print("AttribiteError")
+    print("AttributeError")
 
 try:
     t = T(1)
@@ -39,13 +39,22 @@ try:
 except TypeError:
     print("TypeError")
 
-# Try comma field separator
-T2 = namedtuple("TupComma", "foo,bar")
-t = T2(1, 2)
-print(t)
-print(t.foo, t.bar)
+try:
+    t = T(foo=1)
+except TypeError:
+    print("TypeError")
 
-# Try list of fields
+try:
+    t = T(1, foo=1)
+except TypeError:
+    print("TypeError")
+
+# Try single string
 # Not implemented so far
-#T3 = namedtuple("TupComma", ["foo", "bar"])
+#T3 = namedtuple("TupComma", "foo bar")
 #t = T3(1, 2)
+
+# Try single string with comma field seperator
+# Not implemented so far
+#T2 = namedtuple("TupComma", "foo,bar")
+#t = T2(1, 2)
