@@ -5,10 +5,10 @@ print(spi)
 
 spi = SPI(1, SPI.MASTER)
 spi = SPI(1, SPI.MASTER, baudrate=500000)
-spi = SPI(1, SPI.MASTER, 500000, polarity=1, phase=1, bits=8, firstbit=SPI.MSB, ti=False, crc=None)
+spi = SPI(1, SPI.MASTER, 500000, polarity=1, phase=0, bits=8, firstbit=SPI.MSB, ti=False, crc=None)
 print(spi)
 
-spi.init(SPI.SLAVE)
+spi.init(SPI.SLAVE, phase=1)
 print(spi)
 
 spi.init(SPI.MASTER)
